/***********************************************************************************
* Windows List: Plasmoid to show list of opened windows.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "plasma-windowslist.h"

#include <QPainter>
#include <QFontMetrics>
#include <QSizeF>
#include <QGraphicsLinearLayout>

#include <Plasma/Theme>
#include <Plasma/ToolTipManager>

#include <KLocale>
#include <KRun>
#include <KMenu>
#include <KIcon>
#include <KIconLoader>
#include <KWindowSystem>
#include <KWindowInfo>
#include <NETRootInfo>
#include <kworkspace/kwindowlistmenu.h>

WindowsList::WindowsList(QObject *parent, const QVariantList &args) : Plasma::Applet(parent, args)
{
    setAspectRatioMode(Plasma::ConstrainedSquare);

    int iconSize = IconSize(KIconLoader::Desktop);

    resize((iconSize * 2), (iconSize * 2));
}

void WindowsList::init()
{
    m_icon = new Plasma::IconWidget(KIcon("window-duplicate"), QString(), this);;

    registerAsDragHandle(m_icon);

    QGraphicsLinearLayout *layout = new QGraphicsLinearLayout(this);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);
    layout->addItem(m_icon);

    Plasma::ToolTipContent toolTipData(i18n("Windows list"), i18n("Show list of opened windows"), m_icon->icon().pixmap(IconSize(KIconLoader::Desktop)));

    Plasma::ToolTipManager::self()->setContent(this, toolTipData);

    connect(m_icon, SIGNAL(clicked()), this, SLOT(showMenu()));
}

void WindowsList::showMenu()
{
    KWindowListMenu *menu = new KWindowListMenu();
    QList<WId> windows = KWindowSystem::windows();
    QList< QList<QAction*> > list;
    QAction *action;
    int number;

    action = menu->addAction("Unclutter windows");

    connect(action, SIGNAL(triggered()), menu, SLOT(slotUnclutterWindows()));

    action = menu->addAction("Cascade windows");

    connect(action, SIGNAL(triggered()), menu, SLOT(slotCascadeWindows()));

    for (int i = 0; i <= KWindowSystem::numberOfDesktops(); ++i)
    {
        list.append(QList<QAction*>());
    }

    for (int i = 0; i < windows.count(); ++i)
    {
        KWindowInfo *window = new KWindowInfo(windows.at(i), NET::WMGeometry | NET::WMFrameExtents | NET::WMWindowType | NET::WMDesktop | NET::WMState | NET::XAWMState | NET::WMVisibleName);
        NET::WindowType type = window->windowType(NET::NormalMask | NET::DialogMask | NET::OverrideMask | NET::UtilityMask | NET::DesktopMask | NET::DockMask | NET::TopMenuMask | NET::SplashMask | NET::ToolbarMask | NET::MenuMask);

        if (type == NET::Desktop || type == NET::Dock || type == NET::TopMenu || type == NET::Splash || type == NET::Menu || type == NET::Toolbar || window->hasState(NET::SkipPager))
        {
            windows.removeAt(i);

            --i;

            continue;
        }

        action = new QAction(QIcon(KWindowSystem::icon(windows.at(i))), window->visibleName(), this);
        action->setData((unsigned long long) windows.at(i));

        QFont font = QFont(action->font());

        if (window->isMinimized())
        {
            font.setItalic(true);
        }
        else if (KWindowSystem::activeWindow() == windows.at(i))
        {
            font.setUnderline(true);
            font.setBold(true);
        }

        action->setFont(font);

        number = (window->onAllDesktops()?0:window->desktop());

        QList<QAction*> subList = list.value(number);
        subList.append(action);

        list.replace(number, subList);
    }

    number = 0;

    for (int i = 0; i <= KWindowSystem::numberOfDesktops(); ++i)
    {
        if (list.value(i).isEmpty())
        {
            continue;
        }

        menu->addTitle(i?KWindowSystem::desktopName(i):i18n("On all desktops"));

        for (int j = 0; j < list.value(i).count(); ++j)
        {
            ++number;

            menu->addAction(list.value(i).value(j));
        }
    }

    connect(menu, SIGNAL(triggered(QAction*)), this, SLOT(triggerAction(QAction*)));

    if (number)
    {
        menu->exec(QCursor::pos());
    }

    delete menu;
}

void WindowsList::triggerAction(QAction *action)
{
    if (action->data().type() == QVariant::ULongLong)
    {
        if (KWindowSystem::activeWindow() == action->data().toULongLong())
        {
            KWindowSystem::minimizeWindow(action->data().toULongLong());
        }
        else
        {
            KWindowSystem::activateWindow(action->data().toULongLong());
        }
    }
}

#include "plasma-windowslist.moc"
